(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: iotree.ml,v 1.16.2.1 2011-09-21 13:08:03 fclement Exp $ *)

type ioatree =
   | IOABranch of ioatree * ioatree
   | IOASplit of ioatree * ioatree
   | IOASumSplit of ioatree * ioatree
   | IOALeafs of ioatree array
   | IOALeaf of Unix.sockaddr
   | IOALeafId
;;

type iotree =
   | IOBranch of iotree * iotree
   | IOSplit of iotree * iotree
   | IOSumSplit of iotree * iotree * (IO.ic * IO.oc)
   | IOLeafs of iotree array
   | IOLeaf of (IO.ic * IO.oc)
;;

let rec iotree_of_ioatree tree =
  match tree with
  | IOABranch (t1, t2) ->
    IOBranch
      (iotree_of_ioatree t1,
       iotree_of_ioatree t2)
  | IOASplit (t1, t2) ->
    IOSplit
      (iotree_of_ioatree t1,
       iotree_of_ioatree t2)
  | IOASumSplit (t1, t2) ->
    let final =
      Log.progress (Log.Log 1)
        "starting sum node" in
    let chans = IO.pipe () in
    final " ok!";
    IOSumSplit
      (iotree_of_ioatree t1,
       iotree_of_ioatree t2,
       chans)
  | IOALeafs tv ->
    IOLeafs (Array.map iotree_of_ioatree tv)
  | IOALeaf addr ->
    let final =
      Log.progress (Log.Log 1)
        "trying to connect to %a" IO.dump_sockaddr addr
    in let (ic, oc) = IO.connect addr in
    final " ok!";
    IOLeaf (ic, oc)
  | IOALeafId ->
    let final =
      Log.progress (Log.Log 1) "starting identity node"
    in let (ic, oc) = IO.pipe () in
    final " ok!";
    IOLeaf (ic, oc)
;;

let rec close_tree t =
  match t with
  | IOBranch (t1, t2) | IOSplit (t1, t2) ->
    close_tree t1; close_tree t2
  | IOSumSplit (t1, t2, (ic, oc)) ->
    IO.close_in ic; IO.close_out oc;
    close_tree t1; close_tree t2
  | IOLeafs tv ->
    Array.iter close_tree tv;
  | IOLeaf (ic, oc) ->
    IO.close_in ic; IO.close_out oc;
;;

let rec map_tree tree is =
  match tree with
  | IOBranch (t1, t2) ->
    let f1 x = map_tree t1 x in
    let f2 x = map_tree t2 x in
    Sklstream.in_par (f1, f2) is
  | IOSplit (t1, t2) ->
    let f1 x = map_tree t1 x in
    let f2 x = map_tree t2 x in
    Obj.magic (
      Sklstream.in_prod (f1, f2)
        (Obj.magic is)
    )
  | IOSumSplit (t1, t2, _) ->
    let f1 x = map_tree t1 x in
    let f2 x = map_tree t2 x in
    Sklstream.in_sum (f1, f2) (Obj.magic is)
  | IOLeafs tv ->
    let fv = Array.map map_tree tv in
    Sklstream.on_rails fv is
  | IOLeaf io -> Sklstream.of_to_channel io is
;;

let send_in_tree tree is =
  let rec round_send tree is =
    match tree with
    | IOBranch (t1, t2) ->
      let f1 x = round_send t1 x in
      let f2 x = round_send t2 x in
      Sklstream.in_par (f1, f2) is
    | IOSplit (t1, t2) ->
      let f1 x = round_send t1 x in
      let f2 x = round_send t2 x in
      Obj.magic (
        Sklstream.in_prod (f1, f2)
          (Obj.magic is)
      )
    | IOSumSplit (t1, t2, (_, oc)) ->
      let f1 x = round_send t1 x in
      let f2 x = round_send t2 x in
      let ((s1, s2), cs) = Sklstream.sum_split (Obj.magic is) in
      let s, cs =
        (f1 s1, f2 s2),
        Sklstream.stream_to_channel cs oc in
      Sklstream.sum_combine (s, cs)
    | IOLeafs tv ->
      let fv = Array.map round_send tv in
      Sklstream.on_rails fv is
    | IOLeaf (_, oc) -> Sklstream.stream_to_channel is oc in
  Sklstream.iter ignore (round_send tree is);
  (* close_tree tree *)
;;

let rec receive_in_tree tree =
  match tree with
  | IOBranch (t1, t2) ->
    let r1 = receive_in_tree t1 in
    let r2 = receive_in_tree t2 in
    Sklstream.join (r1, r2)
  | IOSplit (t1, t2) ->
    let r1 = receive_in_tree t1 in
    let r2 = receive_in_tree t2 in
    Obj.magic (Sklstream.combine r1 r2)
  | IOSumSplit (t1, t2, (ic, _)) ->
    let r1 = receive_in_tree t1 in
    let r2 = receive_in_tree t2 in
    let cs = Sklstream.of_channel ic in
    Obj.magic (Sklstream.sum_combine ((r1, r2), cs))
  | IOLeafs tv ->
    let sv = Array.map receive_in_tree tv in
    Sklstream.join_rails sv
  | IOLeaf (ic, _) ->
    Sklstream.of_channel ic
;;
